package com.supervision.knowsub.service.impl;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.lang.Assert;
import cn.hutool.core.lang.UUID;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.supervision.knowsub.domain.UserInfo;
import com.supervision.knowsub.entity.vo.knowledge.KnowledgeFlowRecordResVO;
import com.supervision.knowsub.entity.vo.knowledge.ProcessFlowReqVO;
import com.supervision.knowsub.enums.FlowTypeEnum;
import com.supervision.knowsub.enums.StatusEnum;
import com.supervision.knowsub.exception.BusinessException;
import com.supervision.knowsub.mapper.KnowledgeFlowRecordMapper;
import com.supervision.knowsub.model.*;
import com.supervision.knowsub.service.*;
import com.supervision.knowsub.util.UserUtil;
import com.supervision.knowsub.vo.knowledge.KnowledgeFlowResVO;
import com.supervision.knowsub.vo.knowledge.KnowledgePageResVO;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ObjectUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

@Slf4j
@Service
@RequiredArgsConstructor
public class KnowledgeFlowServiceImpl implements KnowledgeFlowService {

    private final KnowledgeFlowRecordService knowledgeFlowRecordService;

    private final SystemFlowTypeRelationService systemFlowTypeRelationService;

    private final SystemFlowRuleService systemFlowRuleService;

    private final SystemFlowRuleUserService systemFlowRuleUserService;

    private final KnowledgeFlowTodoService knowledgeFlowTodoService;

    private final SystemFlowService systemFlowService;

    private final KnowledgeService knowledgeService;

    private final SystemUserService systemUserService;

    private final SystemDeptService systemDeptService;

    private final KnowledgeSubmitRecordService knowledgeSubmitRecordService;

    @Override
    public IPage<KnowledgeFlowResVO> queryKnowledgeFlowPage(Integer processStatus, String title, String publishDeptId, String baseId,
                                                            String submitDeptId, Instant flowType,
                                                            LocalDate publishDateBegin, LocalDate publishDateEnd, LocalDate submitDateBegin, LocalDate submitDateEnd,
                                                            Integer pageNum, Integer pageSize) {
        Assert.notNull(processStatus, "审批状态不能为空");
        UserInfo user = UserUtil.getUser();
        IPage<KnowledgeFlowResVO> result;
        // 查询待办
        if (1 == processStatus) {
            result = knowledgeFlowRecordService.queryKnowledgeFlowTodoPage(user.getId(), title, publishDeptId, baseId, submitDeptId, flowType,
                    publishDateBegin, publishDateEnd, submitDateBegin, submitDateEnd, new Page<>(pageNum, pageSize));
        }
        // 查询已办
        else if (2 == processStatus) {
            result = knowledgeFlowRecordService.queryKnowledgeFlowCompletePage(user.getId(), title, publishDeptId, baseId, submitDeptId, flowType,
                    publishDateBegin, publishDateEnd, submitDateBegin, submitDateEnd, new Page<>(pageNum, pageSize));
        } else {
            throw new BusinessException("不支持的状态类型");
        }
        // 然后填充一些字段
        List<KnowledgeFlowResVO> records = result.getRecords();
        // 获取所有部门
        Set<String> publishDeptSet = records.stream().map(KnowledgeFlowResVO::getPublishDeptId).collect(Collectors.toSet());
        Set<String> submitDeptSet = records.stream().map(KnowledgeFlowResVO::getSubmitDeptId).collect(Collectors.toSet());
        Collection<String> deptIdSet = CollUtil.union(submitDeptSet, publishDeptSet);
        Map<String, String> deptMap = new HashMap<>();
        if (CollUtil.isNotEmpty(deptIdSet)) {
            deptMap = systemDeptService.listByIds(deptIdSet).stream().collect(Collectors.toMap(SystemDept::getId, SystemDept::getDeptName));
        }
        // 获取所有子库
        Set<String> userIdSet = records.stream().map(KnowledgeFlowResVO::getSubmitUserId).collect(Collectors.toSet());
        Map<String, String> userMap = new HashMap<>();
        if (CollUtil.isNotEmpty(userIdSet)) {
            userMap = systemUserService.listByIds(userIdSet).stream().collect(Collectors.toMap(SystemUser::getId, SystemUser::getUsername));
        }
        for (KnowledgeFlowResVO record : records) {
            record.setSubmitUserName(userMap.getOrDefault(record.getSubmitUserId(), "未知用户"));
            record.setSubmitDeptName(deptMap.getOrDefault(record.getSubmitDeptId(), "未知部门"));
            record.setPublishDeptName(deptMap.getOrDefault(record.getPublishDeptId(), "未知部门"));
        }
        return result;
    }

    /**
     * 提交审批流程
     */
    @Transactional(rollbackFor = Exception.class)
    public void submitFlowProcess(Knowledge knowledge, FlowTypeEnum flowTypeEnum, String userId, String deptId, String remark) {
        String flowInstanceId = UUID.fastUUID().toString();
        // 首先查询到知识对应的子库
        String baseId = knowledge.getBaseId();
        if (StrUtil.isBlank(baseId)) {
            throw new BusinessException("未找到对应的子库");
        }
        // 首先找到对应的审批流
        Optional<SystemFlowTypeRelation> systemFlowTypeRelationOpt = systemFlowTypeRelationService.lambdaQuery()
                .eq(SystemFlowTypeRelation::getBaseId, baseId)
                .eq(SystemFlowTypeRelation::getFlowType, flowTypeEnum.getFlowType())
                .last(" limit 1").oneOpt();
        SystemFlowTypeRelation systemFlowTypeRelation = systemFlowTypeRelationOpt.orElseThrow(() -> new BusinessException("未找到审批流程,请联系管理员配置审批流程"));
        List<SystemFlowRule> roleOrderList = queryFlowRuleWithOrder(systemFlowTypeRelation.getFlowId());
        // 首先获取第一个节点,校验当前用户是否有权限提交
        SystemFlowRule firstRule = roleOrderList.get(0);
        Optional<SystemFlowRuleUser> systemFlowRuleUserOpt = systemFlowRuleUserService.lambdaQuery().eq(SystemFlowRuleUser::getFlowId, firstRule.getFlowId())
                .eq(SystemFlowRuleUser::getUserId, userId).eq(SystemFlowRuleUser::getRuleId, firstRule.getId()).oneOpt();
        if (systemFlowRuleUserOpt.isEmpty()) {
            throw new BusinessException("当前用户没有权限提交:" + flowTypeEnum.getDesc() + " 流程,请联系管理员确认");
        }
        // 然后找到下一个节点
        SystemFlowRule nextRule = roleOrderList.get(1);
        List<SystemFlowRuleUser> nextRuleUserList = systemFlowRuleUserService.lambdaQuery().eq(SystemFlowRuleUser::getFlowId, nextRule.getFlowId()).eq(SystemFlowRuleUser::getRuleId, nextRule.getId()).list();
        if (CollUtil.isEmpty(nextRuleUserList)) {
            throw new BusinessException("审批节点未配置审批用户,流程提交失败");
        }
        // 创建审批记录的对象
        KnowledgeFlowRecord knowledgeFlowRecord = KnowledgeFlowRecord.builder()
                .type(1)
                .knowledgeId(knowledge.getId())
                .flowId(systemFlowTypeRelation.getFlowId())
                // 新建的流程需要赋予一个实例ID
                .flowInstanceId(flowInstanceId)
                .flowType(flowTypeEnum.getFlowType())
                .flowRuleId(firstRule.getId())
                .flowRuleOrder(firstRule.getRuleOrder())
                .processName(flowTypeEnum.getDesc())
                .remark(remark)
                .submitUserId(userId)
                .submitDeptId(deptId)
                .build();
        knowledgeFlowRecordService.save(knowledgeFlowRecord);
        // 只要是发起了审批流程,不管是什么流程,都进行记录发起人
        KnowledgeSubmitRecord knowledgeSubmitRecord = new KnowledgeSubmitRecord();
        knowledgeSubmitRecord.setKnowledgeId(knowledge.getId());
        knowledgeSubmitRecord.setFlowId(systemFlowTypeRelation.getFlowId());
        knowledgeSubmitRecord.setFlowType(systemFlowTypeRelation.getFlowType());
        knowledgeSubmitRecord.setSubmitUserId(userId);
        knowledgeSubmitRecord.setSubmitDeptId(deptId);
        knowledgeSubmitRecordService.save(knowledgeSubmitRecord);

        // 生成下一个节点的TODO
        for (SystemFlowRuleUser todoUser : nextRuleUserList) {
            KnowledgeFlowTodo todo = new KnowledgeFlowTodo();
            todo.setFlowId(nextRule.getFlowId());
            todo.setFlowInstanceId(flowInstanceId);
            todo.setSubmitUserId(flowInstanceId);
            todo.setFlowType(flowTypeEnum.getFlowType());
            todo.setKnowledgeId(knowledge.getId());
            todo.setRuleId(nextRule.getId());
            todo.setTodoUserId(todoUser.getUserId());
            todo.setSubmitUserId(userId);
            todo.insert();
        }
    }

    private List<SystemFlowRule> queryFlowRuleWithOrder(String flowId) {
        List<SystemFlowRule> ruleList = systemFlowRuleService.lambdaQuery().eq(SystemFlowRule::getFlowId, flowId).list();
        if (CollUtil.isEmpty(ruleList)) {
            throw new BusinessException("审批流节点为空,请联系管理员配置审批流节点");
        }
        // 根据rule_order进行升序排序
        List<SystemFlowRule> ruleOrderList = ruleList.stream().sorted(Comparator.comparingInt(SystemFlowRule::getRuleOrder)).toList();
        // 应该最少有两个节点,一个是起始节点,一个是审批节点才行
        if (ruleOrderList.size() < 2) {
            throw new BusinessException("审批流节点数量不足,请联系管理员确认审批流节点是否正确");
        }
        return ruleOrderList;
    }


    @Override
    @Transactional(rollbackFor = Exception.class)
    public void processKnowledgeFlow(ProcessFlowReqVO reqVO) {
        // 首先获取这个审批任务
        KnowledgeFlowTodo knowledgeFlowTodo = knowledgeFlowTodoService.getOptById(reqVO.getTodoId()).orElseThrow(() -> new BusinessException("审批任务不存在"));
        // 找到发起节点
        KnowledgeFlowRecord submitFlowRecord = knowledgeFlowRecordService.lambdaQuery().eq(KnowledgeFlowRecord::getFlowInstanceId, knowledgeFlowTodo.getFlowInstanceId())
                .eq(KnowledgeFlowRecord::getType, 1).orderByDesc(KnowledgeFlowRecord::getCreateTime).last(" limit 1").oneOpt().orElse(new KnowledgeFlowRecord());
        UserInfo user = UserUtil.getUser();
        // 将这个任务对应的TODO删除掉
        knowledgeFlowTodoService.lambdaUpdate().eq(KnowledgeFlowTodo::getFlowInstanceId, knowledgeFlowTodo.getFlowInstanceId()).remove();
        // 然后获取下一个节点
        List<SystemFlowRule> ruleOrderList = queryFlowRuleWithOrder(knowledgeFlowTodo.getFlowId());
        // 从list中找到当前节点
        SystemFlowRule currentRule = null;
        // 然后根据order找到当前节点在ruleOrderList中的索引号(根据ID进行查找)
        int index = 0;
        for (int i = 0; i < ruleOrderList.size(); i++) {
            SystemFlowRule rule = ruleOrderList.get(i);
            if (rule.getId().equals(knowledgeFlowTodo.getRuleId())) {
                index = i;
                currentRule = rule;
            }
        }
        if (ObjectUtils.isEmpty(currentRule)) {
            throw new BusinessException("审批节点不存在");
        }
        if (1 == reqVO.getOperate()) {
            // 如果当前节点的索引号和节点数量-1相等(因为索引从0开始),说明是最后一个节点,这时说明流程走完了
            if ((ruleOrderList.size() - 1) == index) {
                // 这时走流程结束的流程
                KnowledgeFlowRecord newKnowledgeFlowRecord = KnowledgeFlowRecord.builder().type(2).knowledgeId(knowledgeFlowTodo.getKnowledgeId())
                        .flowId(knowledgeFlowTodo.getFlowId()).flowInstanceId(knowledgeFlowTodo.getFlowInstanceId())
                        .flowType(knowledgeFlowTodo.getFlowType()).flowRuleId(currentRule.getId()).flowRuleOrder(currentRule.getRuleOrder())
                        .processName("通过").processStatus(1).processUserId(user.getId())
                        .processDeptId(user.getDeptId()).processTime(LocalDateTime.now()).remark(reqVO.getRemark())
                        .submitUserId(submitFlowRecord.getSubmitUserId()).submitDeptId(submitFlowRecord.getSubmitDeptId())
                        .build();
                knowledgeFlowRecordService.save(newKnowledgeFlowRecord);
                // 将知识的状态进行修改(需要修改为对应的状态,看提交的是什么流程)
                Integer flowType = knowledgeFlowTodo.getFlowType();
                // 如果是知识报送,就置为通过
                if (flowType.equals(FlowTypeEnum.SUBMIT.getFlowType())) {
                    knowledgeService.lambdaUpdate().set(Knowledge::getStatus, StatusEnum.PASS.getStatus()).set(Knowledge::getDraftBelongUserId, null)
                            .eq(Knowledge::getId, reqVO.getKnowledgeId()).update();
                } else if (flowType.equals(FlowTypeEnum.RECALL.getFlowType())) {
                    // 如果是撤回,就置为草稿,并且草稿所属人应该是提交流程的人
                    knowledgeService.lambdaUpdate().set(Knowledge::getStatus, StatusEnum.DRAFT.getStatus()).set(Knowledge::getDraftBelongUserId, submitFlowRecord.getSubmitUserId())
                            .eq(Knowledge::getId, reqVO.getKnowledgeId()).update();
                } else if (flowType.equals(FlowTypeEnum.DELETE.getFlowType())) {
                    knowledgeService.lambdaUpdate().set(Knowledge::getStatus, StatusEnum.DELETE.getStatus()).set(Knowledge::getDraftBelongUserId, null)
                            .eq(Knowledge::getId, reqVO.getKnowledgeId()).update();
                } else {
                    throw new BusinessException("未知的流程类型");
                }
            } else {
                SystemFlowRule nextRule;
                try {
                    nextRule = ruleOrderList.get(index + 1);
                } catch (IndexOutOfBoundsException e) {
                    throw new BusinessException("未找到下一个审批节点,请联系管理员解决");
                }
                List<SystemFlowRuleUser> nextRuleUserList = systemFlowRuleUserService.lambdaQuery().eq(SystemFlowRuleUser::getFlowId, nextRule.getFlowId())
                        .eq(SystemFlowRuleUser::getRuleId, nextRule.getId()).list();
                if (CollUtil.isEmpty(nextRuleUserList)) {
                    throw new BusinessException("下级审批节点未配置审批用户,审批失败,请联系管理员解决");
                }
                // 保存这个节点的审批记录
                KnowledgeFlowRecord newKnowledgeFlowRecord = KnowledgeFlowRecord.builder().type(2).knowledgeId(knowledgeFlowTodo.getKnowledgeId())
                        .flowId(knowledgeFlowTodo.getFlowId()).flowInstanceId(knowledgeFlowTodo.getFlowInstanceId())
                        .flowType(knowledgeFlowTodo.getFlowType()).flowRuleId(currentRule.getId()).flowRuleOrder(currentRule.getRuleOrder())
                        .processName("通过").processStatus(1).processUserId(user.getId())
                        .processDeptId(user.getDeptId()).processTime(LocalDateTime.now()).remark(reqVO.getRemark())
                        .submitUserId(submitFlowRecord.getSubmitUserId()).submitDeptId(submitFlowRecord.getSubmitDeptId())
                        .build();
                knowledgeFlowRecordService.save(newKnowledgeFlowRecord);
                // 生成下一个节点的TODO
                for (SystemFlowRuleUser todoUser : nextRuleUserList) {
                    KnowledgeFlowTodo todo = new KnowledgeFlowTodo();
                    todo.setFlowId(nextRule.getFlowId());
                    todo.setKnowledgeId(knowledgeFlowTodo.getKnowledgeId());
                    todo.setFlowInstanceId(knowledgeFlowTodo.getFlowInstanceId());
                    todo.setTodoUserId(todoUser.getUserId());
                    todo.setSubmitUserId(submitFlowRecord.getSubmitUserId());
                    todo.setFlowType(knowledgeFlowTodo.getFlowType());
                    todo.setRuleId(nextRule.getId());
                    todo.insert();
                }
            }
        }
        // 驳回流程
        else if (2 == reqVO.getOperate()) {
            KnowledgeFlowRecord newKnowledgeFlowRecord = KnowledgeFlowRecord.builder().type(2)
                    .knowledgeId(knowledgeFlowTodo.getKnowledgeId())
                    .flowId(knowledgeFlowTodo.getFlowId())
                    .flowType(knowledgeFlowTodo.getFlowType())
                    .flowRuleId(currentRule.getId())
                    .flowInstanceId(knowledgeFlowTodo.getFlowInstanceId())
                    .flowRuleOrder(currentRule.getRuleOrder())
                    .processName("驳回")
                    .processStatus(2)
                    .processUserId(user.getId())
                    .processDeptId(user.getDeptId())
                    .processTime(LocalDateTime.now())
                    .remark(reqVO.getRemark())
                    .submitUserId(submitFlowRecord.getSubmitUserId())
                    .submitDeptId(submitFlowRecord.getSubmitDeptId()).build();
            knowledgeFlowRecordService.save(newKnowledgeFlowRecord);
            // 驳回,需要分辨驳回的状态是什么
            if (knowledgeFlowTodo.getFlowType().equals(FlowTypeEnum.SUBMIT.getFlowType())) {
                // 如果是知识报送,就置为驳回(审批)
                knowledgeService.lambdaUpdate().set(Knowledge::getStatus, StatusEnum.RECALL_APPROVAL_SUBMIT.getStatus()).set(Knowledge::getDraftBelongUserId, knowledgeFlowTodo.getSubmitUserId())
                        .eq(Knowledge::getId, reqVO.getKnowledgeId()).update();
            } else if (knowledgeFlowTodo.getFlowType().equals(FlowTypeEnum.RECALL.getFlowType())) {
                // 如果是撤回,就置为驳回(撤回)
                knowledgeService.lambdaUpdate().set(Knowledge::getStatus, StatusEnum.RECALL_APPROVAL_RECALL.getStatus()).set(Knowledge::getDraftBelongUserId, submitFlowRecord.getSubmitUserId())
                        .eq(Knowledge::getId, reqVO.getKnowledgeId()).update();
            } else if (knowledgeFlowTodo.getFlowType().equals(FlowTypeEnum.DELETE.getFlowType())) {
                // 如果是删除,就置为驳回(删除)
                knowledgeService.lambdaUpdate().set(Knowledge::getStatus, StatusEnum.RECALL_APPROVAL_DELETE.getStatus()).set(Knowledge::getDraftBelongUserId, submitFlowRecord.getSubmitUserId())
                        .eq(Knowledge::getId, reqVO.getKnowledgeId()).update();
            } else {
                throw new BusinessException("未知的流程类型");
            }
        } else {
            throw new BusinessException("暂不支持的审批类型");
        }
    }

    @Override
    public List<KnowledgeFlowRecordResVO> queryKnowledgeFlowRecord(String knowledgeId, String flowInstanceId) {
        List<KnowledgeFlowRecord> list = knowledgeFlowRecordService.lambdaQuery().eq(KnowledgeFlowRecord::getKnowledgeId, knowledgeId)
                .eq(StrUtil.isNotBlank(flowInstanceId), KnowledgeFlowRecord::getFlowInstanceId, flowInstanceId).list();

        // 根据时间进行排序
        List<KnowledgeFlowRecordResVO> result = list.stream().map(knowledgeFlowRecord -> {
            KnowledgeFlowRecordResVO node = new KnowledgeFlowRecordResVO();
            node.setKnowledgeId(knowledgeFlowRecord.getKnowledgeId());
            node.setProcessName(knowledgeFlowRecord.getProcessName());
            node.setRemark(knowledgeFlowRecord.getRemark());
            node.setProcessOrder(knowledgeFlowRecord.getFlowRuleOrder());
            // 判断是不是发起的,如果是发起的,那么process的是发起人.否则是审批人
            if (1 == knowledgeFlowRecord.getType()) {
                node.setProcessTime(knowledgeFlowRecord.getCreateTime());
                node.setProcessUserId(knowledgeFlowRecord.getSubmitUserId());
                node.setProcessDeptId(knowledgeFlowRecord.getSubmitDeptId());
            } else {
                node.setProcessTime(knowledgeFlowRecord.getProcessTime());
                node.setProcessUserId(knowledgeFlowRecord.getProcessUserId());
                node.setProcessDeptId(knowledgeFlowRecord.getProcessDeptId());

            }
            return node;
        }).sorted(Comparator.comparing(KnowledgeFlowRecordResVO::getProcessTime).reversed()).toList();
        // 然后把人名和部门名进行转换
        Set<String> userIdSet = result.stream().map(KnowledgeFlowRecordResVO::getProcessUserId).collect(Collectors.toSet());
        List<SystemUser> systemUsers = new ArrayList<>();
        if (CollUtil.isNotEmpty(userIdSet)) {
            systemUsers = systemUserService.listByIds(userIdSet);
        }
        Map<String, String> userNameMap = systemUsers.stream().collect(Collectors.toMap(SystemUser::getId, SystemUser::getUsername));
        Set<String> deptIdSet = result.stream().map(KnowledgeFlowRecordResVO::getProcessDeptId).collect(Collectors.toSet());
        List<SystemDept> systemDeptList = new ArrayList<>();
        if (CollUtil.isNotEmpty(deptIdSet)) {
            systemDeptList = systemDeptService.listByIds(deptIdSet);
        }

        Map<String, String> deptNameMap = systemDeptList.stream().collect(Collectors.toMap(SystemDept::getId, SystemDept::getDeptName));
        result.forEach(node -> {
            node.setProcessUserName(userNameMap.getOrDefault(node.getProcessUserId(), "未知用户"));
            node.setProcessDeptName(deptNameMap.getOrDefault(node.getProcessDeptId(), "未知部门"));
        });
        return result;
    }
}
